#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

const size_t BLOCK = 16384;
/* don't copy setuid/setgid/sticky bits */
const mode_t PERM_MASK = S_IRWXU | S_IRWXG | S_IRWXO;

const char *prog;

int
perform_copy(int src, mode_t perms, char *const *paths, size_t num_dests)
{
	struct stat	st;
	off_t	total = 0;
	char	*buf = 0;
	int	*fds = 0;
	int	status;
	bool	open_fail;

	if (!(buf = malloc(BLOCK))) {
		fprintf(stderr, "%s: out of memory\n", prog);
		status = -1;
		goto end;
	}

	fds = malloc(sizeof(int) * num_dests);
	if (!fds) {
		fprintf(stderr, "%s: out of memory\n", prog);
		status = -1;
		goto end;
	}
	for (size_t i = 0; i < num_dests; i++)
		fds[i] = -1;

	if (fstat(src, &st) == -1) {
		fprintf(stderr, "%s: stat on source: %s\n",
		    prog, strerror(errno));
		status = 1;
		goto end;
	}

	if (!perms)
		/* copy perms from source by default */
		perms = st.st_mode & PERM_MASK;

	open_fail = false;
	for (size_t i = 0; i < num_dests; i++) {
		fds[i] = open(paths[i], O_WRONLY | O_CREAT | O_EXCL, perms);
		if (fds[i] == -1) {
			fprintf(stderr, "%s: failed to create `%s': %s\n",
			    prog, paths[i], strerror(errno));
			open_fail = true;
			break;
		}
	}
	if (open_fail) {
		/* remove any created files */
		for (size_t i = 0; i < num_dests; i++) {
			if (fds[i] <= 2)
				break;
			unlink(paths[i]);
		}
		status = 1;
		goto end;
	}

	for (;;) {
		ssize_t	bytes;
		size_t	size;
		size_t	written;

		/* read a block */
		size = 0;
		while (size < BLOCK) {
			bytes = read(src, buf + size, BLOCK - size);
			if (bytes < 0) {
				if (size)
					break;
				fprintf(stderr, "%s: read error: %s\n",
				    prog, strerror(errno));
				status = 1;
				goto end;
			} else if (!bytes)
				break;
			size += bytes;
		}
		/* eof */
		if (!size)
			break;

		/* write the block to all dests */
		for (size_t i = 0; i < num_dests; i++) {
			if (fds[i] == -1)
				/* skip failed writes */
				continue;

			/* write the block */
			written = 0;
			while (written < size) {
				bytes = write(fds[i], buf + written,
				    size - written);
				if (bytes < 0) {
					fprintf(stderr,
					    "%s: write error on %s: %s\n",
					    prog, paths[i], strerror(errno));
					close(fds[i]);
					fds[i] = -1;
					break;
				}
				written += bytes;
			}
		}

		total += size;
		printf("%2u%%\r", (unsigned)(100 * total / st.st_size));
		fflush(stdout);
	}

	status = 0;
end:
	if (fds) {
		for (size_t i = 0; i < num_dests; i++)
			if (fds[i] != -1)
				close(fds[i]);
		free(fds);
	}
	if (buf) free(buf);
	return status;
}

int
main(int argc, char **argv)
{
	struct stat	st;
	const char	*name;
	size_t		name_len;
	char		**paths = 0;
	size_t		src_len;
	unsigned	dests;
	int		src_fd = -1;
	int		status;
	mode_t		perms; /* prob. 16 bits */

	prog = *argv;
	if (argc < 3) {
		fprintf(stderr, "Usage: %s FILE DEST...\n", prog);
		status = 1;
		goto end;
	}

	/* get name and name length */
	src_len = strlen(argv[1]);
	name_len = 1;
	name = argv[1] + src_len - 1;
	for (;;) {
		if (*name == '/') {
			name++;
			name_len--;
			break;
		} else if (name == argv[1])
			break;
		name--;
		name_len++;
	}

	/* open and check source */
	if ((src_fd = open(argv[1], O_RDONLY)) == -1) {
		fprintf(stderr, "%s: opening source: %s\n",
		    prog, strerror(errno));
		status = 1;
		goto end;
	} else if (fstat(src_fd, &st) == -1) {
		fprintf(stderr, "%s: checking source: %s\n",
		    prog, strerror(errno));
		status = 1;
		goto end;
	} else if (!S_ISREG(st.st_mode)) {
		fprintf(stderr, "%s: source is not a regular file\n", prog);
		status = 1;
		goto end;
	}
	perms = st.st_mode & PERM_MASK;

	/* allocate memory for string array */
	dests = argc - 2;
	paths = malloc(sizeof(char *) * dests);
	if (!paths) {
		fprintf(stderr, "%s: out of memory\n", prog);
		status = 1;
		goto end;
	}
	for (size_t i = 0; i < dests; i++)
		paths[i] = 0;

	/* create/copy destination pathnames */
	for (size_t i = 0; i < dests; i++) {
		size_t	    len;
		const char  *path = argv[i+2];
		bool	    is_dir = false;
		if (stat(path, &st) == -1) {
			if (errno != ENOENT) {
				fprintf(stderr, "%s: `%s': %s\n",
				    prog, argv[i+2], strerror(errno));
				status = 1;
				goto end;
			}
		} else if (S_ISDIR(st.st_mode))
			is_dir = true;

		len = strlen(path);
		if (is_dir) {
			char	*pos;
			bool	need_slash;
			need_slash = path[len-1] != '/';
			paths[i] = malloc(len + name_len + need_slash + 1);
			if (!paths[i]) {
				fprintf(stderr, "%s: out of memory\n", prog);
				status = 1;
				goto end;
			}

			pos = paths[i];
			memcpy(pos, path, len);
			pos += len;

			if (need_slash)
				*pos++ = '/';

			memcpy(pos, name, name_len + 1);
		} else {
			paths[i] = malloc(len + 1);
			memcpy(paths[i], path, len + 1);
		}
	}

	if (perform_copy(src_fd, perms, paths, dests) == -1) {
		status = 1;
		goto end;
	}

	status = 0;
end:
	if (paths) {
		for (size_t i = 0; i < dests; i++)
			if (paths[i])
				free(paths[i]);
		free(paths);
	}
	if (src_fd != -1)
		close(src_fd);

	return status;
}
